﻿using System;
using System.Collections.Generic;
using System.Text;
using SharpMap.Geometries;
using System.Xml;
using Ogc.Filter;
using SharpMap.Data;

namespace Ogc.Sld
{
    public class Rule
    {
        Symbolizer symbolizer;
        IFilter filter;

        double minZoom;
        double maxZoom;

        Rule elseRule;
        public Rule ElseRule
        {
            get { return elseRule; }
            set { elseRule = value; }
        }

        public static Rule[] FromSldFile(string sldfile, IWellKnownSymbolGetter symbolGetter)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(sldfile);
            Rule prev = null;
            List<Rule> ret = new List<Rule>();
            foreach (XmlNode i in doc.SelectNodes("//Rule"))
            {
                bool isElse = false;
                Rule r = BuildStyleRule(i, symbolGetter, out isElse);
                if (!isElse)
                    ret.Add(r);
                else
                {
                    if (prev == null)
                        throw new Exception("文件格式有误，ElseFilter之前必须有Filter。");
                    prev.elseRule = r;
                }
                prev = r;
            }
            return ret.ToArray();
        }

        private static Rule BuildStyleRule(XmlNode xmlnode, IWellKnownSymbolGetter symbolGetter, out bool isElse)
        {
            Rule ret = new Rule(BuildSymbolizer(xmlnode, symbolGetter), BuildFilter(xmlnode, out isElse));

            XmlNode n = xmlnode.SelectSingleNode("MinScaleDenominator");
            if (n != null)
                ret.minZoom = double.Parse(n.InnerText);
            else
                ret.minZoom = double.NaN;
            n = xmlnode.SelectSingleNode("MaxScaleDenominator");
            if (n != null)
                ret.maxZoom = double.Parse(n.InnerText);
            else
                ret.maxZoom = double.NaN;
            return ret;
        }

        private static Symbolizer BuildSymbolizer(XmlNode xmlnode, IWellKnownSymbolGetter symbolGetter)
        {
            return new SymbolizerBuilder(symbolGetter).Build(xmlnode);
        }

        private static IFilter BuildFilter(XmlNode xmlnode, out bool isElse)
        {
            FilterBuilder builder = new FilterBuilder();

            XmlNode sn = xmlnode.SelectSingleNode("Filter");
            if (sn != null)
            {
                isElse = false;
                return builder.BuildFilter(sn);
            }
            sn = xmlnode.SelectSingleNode("ElseFilter");
            if (sn != null)
            {
                isElse = true;
                return builder.BuildFilter(sn);
            }
            isElse = false;
            return null;
        }

        public Rule(Symbolizer symbolizer, IFilter tester)
        {
            this.symbolizer = symbolizer;
            this.filter = tester;
        }

        public bool Test(SharpMap.Data.FeatureDataRow dr, double zoom, out Symbolizer symbolizer)
        {
            if (TestZoom(zoom))
            {
                if (filter == null || filter.Test(new FeatureDataRowGetter(dr)))
                {
                    symbolizer = this.symbolizer;
                    return true;
                }

                if (elseRule != null)
                    return elseRule.Test(dr, zoom, out symbolizer);
            }
            symbolizer = null;
            return false;
        }

        private bool TestZoom(double zoom)
        {
            if (minZoom != double.NaN && minZoom > zoom)
                return false;
            if (maxZoom != double.NaN && maxZoom < zoom)
                return false;
            return true;
        }

    }

    class FeatureDataRowGetter : IPropertyGetter
    {
        FeatureDataRow dr;

        public FeatureDataRowGetter(FeatureDataRow dr)
        {
            this.dr = dr;
        }

        #region IPropertyGetter 成员

        public object GetProperty(string name)
        {
            return dr[name];
        }

        public Geometry GetGeometryProperty(string name)
        {
            return dr.Geometry;
        }

        #endregion
    }
}
