﻿using System;
using System.Collections.Generic;
using System.Text;
using SharpMap.Styles;
using System.Xml;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace Ogc.Sld
{
    class StyleBuilder
    {
        IWellKnownSymbolGetter symbolGetter;

        public StyleBuilder(IWellKnownSymbolGetter symbolGetter)
        {
            this.symbolGetter = symbolGetter;
        }

        internal SharpMap.Styles.IStyle BuildLineSymbolStyle(System.Xml.XmlNode sn)
        {
            return ParseLineSymbolStyle(sn);
        }

        internal SharpMap.Styles.IStyle BuildPointSymbolStyle(System.Xml.XmlNode sn)
        {
            return ParsePointSymbolStyle(sn);
        }

        internal SharpMap.Styles.IStyle BuildPolygonSymbolStyle(System.Xml.XmlNode sn)
        {
            return ParsePolygonSymbolStyle(sn);
        }

        private VectorStyle ParseLineSymbolStyle(XmlNode sn)
        {
            VectorStyle ret = new VectorStyle();
            ret.Line = GetStroke(sn);
            return ret;
        }

        private static Pen GetStroke(XmlNode sn)
        {
            Color clr = Color.Black;
            XmlNode n = sn.SelectSingleNode("Stroke/CssParameter[@name='stroke']");
            if (n != null)
                clr = ParseColorFromCSS(n.InnerText);
            int width = 1;
            n = sn.SelectSingleNode("Stroke/CssParameter[@name='stroke-width']");
            if (n != null)
                width = int.Parse(n.InnerText);
            Pen stroke = new Pen(clr, width);
            return stroke;
        }

        private static Color ParseColorFromCSS(string css)
        {
            return Color.FromArgb(Convert.ToInt32(css.Substring(1, 2), 16)
                , Convert.ToInt32(css.Substring(3, 2), 16)
                , Convert.ToInt32(css.Substring(5, 2), 16));
        }

        private VectorStyle ParsePointSymbolStyle(XmlNode sn)
        {
            VectorStyle ret = new VectorStyle();
            float size = 6;
            float rotation = 0;

            XmlNode n = sn.SelectSingleNode("Graphic/Size");
            if (n != null)
                size = float.Parse(n.InnerText);

            n = sn.SelectSingleNode("Graphic/Rotation");
            if (n != null)
                rotation = float.Parse(n.InnerText);

            //Set WellKnownName Mark
            Bitmap bmp = null;
            n = sn.SelectSingleNode("Graphic/Mark/WellKnownName");
            if (n != null)
            {
                string wkn = n.InnerText;
                n = sn.SelectSingleNode("Graphic/Mark");
                bmp = GetWellKnownBitmap(wkn, size, ParseFill(n), GetStroke(n));
            }

            //or set ExternalGraphic
            if (bmp == null)
            {
                n = sn.SelectSingleNode("Graphic/ExternalGraphic/OnlineResource@xlink:href");
            }

            if (bmp != null)
                ret.Symbol = bmp;

            ret.SymbolOffset = new PointF(-size, -size);
            ret.SymbolRotation = rotation;
            return ret;
        }

        private Bitmap GetWellKnownBitmap(string wkn, float size, Brush brush, Pen pen)
        {
            GraphicsPath gp = symbolGetter.GetSymbol(wkn);

            if (gp == null)
                return null;

            using (gp)
            {
                Matrix m = new Matrix();
                m.Scale(size, size);
                gp.Transform(m);

                Bitmap ret = new Bitmap((int)size, (int)size);
                using (Graphics g = Graphics.FromImage(ret))
                {
                    if (brush != null)
                        g.FillPath(brush, gp);
                    if (pen != null)
                        g.DrawPath(pen, gp);
                }
                return ret;
            }
        }
        
        private VectorStyle ParsePolygonSymbolStyle(XmlNode sn)
        {
            VectorStyle ret = new VectorStyle();

            Brush b = ParseFill(sn);
            if (b != null)
                ret.Fill = b;

            return ret;
        }

        private static Brush ParseFill(XmlNode sn)
        {
            Brush b = null;
            XmlNode n = sn.SelectSingleNode("Fill/CssParameter[@name='fill']");
            if (n != null)
                b = new SolidBrush(ParseColorFromCSS(n.InnerText));
            return b;
        }
    }

    public interface IWellKnownSymbolGetter
    {
        GraphicsPath GetSymbol(string wkn);
    }
}
