﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Drawing;
using System.Drawing.Drawing2D;
using Ogc.Sld.Symbolizers;

namespace Ogc.Sld
{
    class SymbolizerBuilder
    {
        static readonly string[] SymbolizerNames = new string[] {
            "LineSymbolizer","PolygonSymbolizer","PointSymbolizer","TextSymbolizer","RasterSymbolizer"};
        IWellKnownSymbolGetter symbolGetter;

        public SymbolizerBuilder(IWellKnownSymbolGetter symbolGetter)
        {
            this.symbolGetter = symbolGetter;
        }

        public Symbolizer Build(XmlNode xmlnode)
        {
            foreach (string i in SymbolizerNames)
            {
                XmlNode n = xmlnode.SelectSingleNode(i);
                if (n != null)
                    return ((ISymbolizerFactory)Activator.CreateInstance(Type.GetType("Ogc.Sld." + i + "Factory")))
                        .FromXmlNode(n, new ParseContext(symbolGetter));
            }
            throw new NotImplementedException();
        }

    }

    public interface IWellKnownSymbolGetter
    {
        GraphicsPath GetSymbol(string wkn);
    }

    interface ISymbolizerFactory
    {
        Symbolizer FromXmlNode(XmlNode xmlnode,ParseContext context);
    }

    class LineSymbolizerFactory : ISymbolizerFactory
    {
        #region ISymbolizerFactory 成员

        public Symbolizer FromXmlNode(XmlNode xmlnode, ParseContext context)
        {
            LineSymbolizer ret = new LineSymbolizer();
            ret.Stroke = context.ParseStroke(xmlnode);
            return ret;
        }

        #endregion
    }

    class PolygonSymbolizerFactory : ISymbolizerFactory
    {
        #region ISymbolizerFactory 成员

        public Symbolizer FromXmlNode(XmlNode xmlnode, ParseContext context)
        {
            PolygonSymbolizer ret = new PolygonSymbolizer();

            Brush b = context.ParseFill(xmlnode);
            if (b != null)
                ret.Fill = b;

            return ret;
        }

        #endregion
    }

    class PointSymbolizerFactory : ISymbolizerFactory
    {
        #region ISymbolizerFactory 成员

        public Symbolizer FromXmlNode(XmlNode xmlnode, ParseContext context)
        {
            PointSymbolizer ret = new PointSymbolizer();
            float size = 6;
            float rotation = 0;

            XmlNode n = xmlnode.SelectSingleNode("Graphic/Size");
            if (n != null)
                size = float.Parse(n.InnerText);

            n = xmlnode.SelectSingleNode("Graphic/Rotation");
            if (n != null)
                rotation = float.Parse(n.InnerText);

            //Set WellKnownName Mark
            Bitmap bmp = null;
            n = xmlnode.SelectSingleNode("Graphic/Mark/WellKnownName");
            if (n != null)
            {
                string wkn = n.InnerText;
                n = xmlnode.SelectSingleNode("Graphic/Mark");
                bmp = context.GetWellKnownBitmap(wkn, size, context.ParseFill(n), context.ParseStroke(n));
            }

            //or set ExternalGraphic
            if (bmp == null)
            {
                n = xmlnode.SelectSingleNode("Graphic/ExternalGraphic/OnlineResource@xlink:href");
            }

            if (bmp != null)
                ret.Symbol = bmp;

            ret.Size = new SizeF(-size, -size);
            ret.Rotation = rotation;
            return ret;
        }

        #endregion
    }

    class TextSymbolizerFactory : ISymbolizerFactory
    {
        #region ISymbolizerFactory 成员

        public Symbolizer FromXmlNode(XmlNode xmlnode, ParseContext context)
        {
            TextSymbolizer ret = new TextSymbolizer(xmlnode.SelectSingleNode("Label").InnerText);
            string family = "";
            float size = 21;
            FontStyle style = FontStyle.Regular;

            XmlNode n = xmlnode.SelectSingleNode("Font/CssParameter[@name='font-family']");
            if (n != null)
                family = n.InnerText;
            n = xmlnode.SelectSingleNode("Font/CssParameter[@name='font-size']");
            if (n != null)
                size = float.Parse(n.InnerText);
            n = xmlnode.SelectSingleNode("Font/CssParameter[@name='font-weight']");
            if (n != null)
                style = (n.InnerText == "bold" ? FontStyle.Bold : FontStyle.Regular);

            ret.Font = new Font(family, size, style);
            return ret;
        }

        #endregion
    }

    class RasterSymbolizerFactory : ISymbolizerFactory
    {
        #region ISymbolizerFactory 成员

        public Symbolizer FromXmlNode(XmlNode xmlnode, ParseContext context)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        #endregion
    }

    class ParseContext
    {
        IWellKnownSymbolGetter symbolGetter;

        public ParseContext(IWellKnownSymbolGetter symbolGetter)
        {
            this.symbolGetter = symbolGetter;
        }

        public Pen ParseStroke(XmlNode sn)
        {
            Color clr = Color.Black;
            XmlNode n = sn.SelectSingleNode("Stroke/CssParameter[@name='stroke']");
            if (n != null)
                clr = ParseColorFromCSS(n.InnerText);
            int width = 1;
            n = sn.SelectSingleNode("Stroke/CssParameter[@name='stroke-width']");
            if (n != null)
                width = int.Parse(n.InnerText);
            Pen stroke = new Pen(clr, width);
            return stroke;
        }

        public static Color ParseColorFromCSS(string css)
        {
            return Color.FromArgb(Convert.ToInt32(css.Substring(1, 2), 16)
                , Convert.ToInt32(css.Substring(3, 2), 16)
                , Convert.ToInt32(css.Substring(5, 2), 16));
        }

        public Bitmap GetWellKnownBitmap(string wkn, float size, Brush brush, Pen pen)
        {
            GraphicsPath gp = symbolGetter.GetSymbol(wkn);

            if (gp == null)
                return null;

            using (gp)
            {
                Matrix m = new Matrix();
                m.Scale(size, size);
                gp.Transform(m);

                Bitmap ret = new Bitmap((int)size, (int)size);
                using (Graphics g = Graphics.FromImage(ret))
                {
                    if (brush != null)
                        g.FillPath(brush, gp);
                    if (pen != null)
                        g.DrawPath(pen, gp);
                }
                return ret;
            }
        }

        public Brush ParseFill(XmlNode sn)
        {
            Brush b = null;
            XmlNode n = sn.SelectSingleNode("Fill/CssParameter[@name='fill']");
            if (n != null)
                b = new SolidBrush(ParseColorFromCSS(n.InnerText));
            return b;
        }
    }
}
