﻿using System;
using System.Data;
using System.Configuration;
using System.IO;
using SharpMap.CoordinateSystems;
using SharpMap.CoordinateSystems.Transformations;
using SharpMap.Styles;
using System.Xml;
using System.Drawing;
using System.Drawing.Drawing2D;
using SharpMap.Geometries;
using SharpMap;
using SharpMap.Layers;
using SharpMap.Data.Providers;
using System.Collections.Generic;
using System.Web;
using Ogc.SharpMapExtentions;

/// <summary>
/// GetMapProcessor 的摘要说明
/// </summary>
public class GetMapProcessor
{
    HttpContext context;
    public GetMapProcessor(HttpContext context)
    {
        this.context = context;
    }

    //部分代码抄袭了SharpMap
    internal void Process()
    {
        string layers = context.Request["LAYERS"];
        string styles = context.Request["STYLES"];
        string srs = context.Request["SRS"];
        BoundingBox bbox = ParseBBox(context.Request["BBOX"]);
        System.Drawing.Imaging.ImageCodecInfo imageEncoder = GetEncoderInfo(context.Request.Params["FORMAT"]);
        int width = int.Parse(context.Request["WIDTH"]);
        int height = int.Parse(context.Request["HEIGHT"]);

        string[] layerNames = layers.Split(',');
        if (layerNames.Length == 0)
            throw new InvalidDataException("Layers设置不正确");

        string[] styleNames = styles.Split(',');
        if (!string.IsNullOrEmpty(styles) && styleNames.Length != layerNames.Length)
            throw new InvalidDataException("Styles设置不正确");

        Map map = new Map(new System.Drawing.Size(width, height));
        BuildMapLayers(srs, layerNames, styleNames, map);

        map.PixelAspectRatio = ((double)width / (double)height) / (bbox.Width / bbox.Height);
        map.Center = bbox.GetCentroid();
        map.Zoom = bbox.Width;

        using (System.Drawing.Image img = map.GetMap())
        {
            //Png can't stream directy. Going through A memorystream instead
            System.IO.MemoryStream MS = new System.IO.MemoryStream();
            img.Save(MS, imageEncoder, null);
            img.Dispose();
            byte[] buffer = MS.ToArray();
            context.Response.ContentType = imageEncoder.MimeType;
            context.Response.OutputStream.Write(buffer, 0, buffer.Length);
        }
    }

    private void BuildMapLayers(string srs, string[] layerNames, string[] styleNames, Map map)
    {
        for (int i = 0; i < layerNames.Length; ++i)
        {
            string styleName = (styleNames.Length > i ? styleNames[i] : "");
            string layerName = layerNames[i];

            if (string.IsNullOrEmpty(styleName))
                styleName = GetDefaultStyleName(layerName);

            map.Layers.Add(CreateLayer(srs, layerName, Config.GetSldFile(styleName), new WellKnownSymbol()));
        }
    }

    private ILayer CreateLayer(string srs, string layerName, string sldfile, WellKnownSymbol wellKnownSymbol)
    {
        string dsf = Config.GetDataProviderFile(layerName);
        XmlDocument doc = new XmlDocument();
        doc.Load(dsf);
        XmlNode n = doc.SelectSingleNode("/Data/EsriShape");
        IProvider pvd = null;
        ICoordinateTransformation trans = null;
        if (n != null)
        {
            ShapeFile sf = new ShapeFile(Config.GetEsriShapeFile(n.InnerText));
            if (sf.CoordinateSystem != null)
                trans = new CoordinateTransformationFactory()
                    .CreateFromCoordinateSystems(sf.CoordinateSystem, GetCoordinateSystemBySRS(srs));
            pvd = sf;
        }
        n = doc.SelectSingleNode("/Data/WFS");
        if (n != null)
        {
            string geo = n.SelectSingleNode("GeoProperty").InnerText;
            string srcSrs = n.SelectSingleNode("SRS").InnerText;
            pvd = new WfsProvider(n.SelectSingleNode("Uri").InnerText
                , n.SelectSingleNode("Name").InnerText
                , new string[] { geo }, geo, srcSrs, BuildBBox((XmlElement)n.SelectSingleNode("BoundingBox")));
            if (srcSrs != srs)
                trans = new CoordinateTransformationFactory()
                        .CreateFromCoordinateSystems(GetCoordinateSystemBySRS(srcSrs), GetCoordinateSystemBySRS(srs));
        }
        if (pvd == null)
            throw new NotImplementedException();

        StyledLayer ret = new StyledLayer();
        ret.Rules = Ogc.Sld.Rule.FromSldFile(sldfile, wellKnownSymbol);
        ret.DataSource = pvd;
        ret.CoordinateTransformation = trans;
        return ret;
    }

    private string GetDefaultStyleName(string layerName)
    {
        string lf = Config.GetWmsLayerFile(layerName);
        XmlDocument doc = new XmlDocument();
        doc.Load(lf);
        foreach (XmlNode i in doc.SelectNodes("Layer/Style"))
        {
            return i.InnerText;
        }
        throw new InvalidOperationException();
    }

    private BoundingBox BuildBBox(XmlElement xmlNode)
    {
        return new BoundingBox(double.Parse(xmlNode.GetAttribute("minx"))
            , double.Parse(xmlNode.GetAttribute("miny"))
            , double.Parse(xmlNode.GetAttribute("maxx"))
            , double.Parse(xmlNode.GetAttribute("maxy")));
    }

    private ICoordinateSystem GetCoordinateSystemBySRS(string srs)
    {
        string[] ss = srs.Split(':');
        if (ss.Length != 2)
            throw new InvalidDataException();

        foreach (string i in File.ReadAllLines(Config.GetSridFile()))
        {
            if (i.StartsWith(ss[1]))
                return new SharpMap.CoordinateSystems.CoordinateSystemFactory()
                    .CreateFromWkt(i.Substring(i.IndexOf(';') + 1));
        }

        throw new NotImplementedException("还不支持这种SRS" + srs);
    }


    //部分代码抄袭了SharpMap
    private static System.Drawing.Imaging.ImageCodecInfo GetEncoderInfo(String mimeType)
    {
        foreach (System.Drawing.Imaging.ImageCodecInfo encoder in System.Drawing.Imaging.ImageCodecInfo.GetImageEncoders())
            if (encoder.MimeType == mimeType)
                return encoder;
        return null;
    }

    private BoundingBox ParseBBox(string p)
    {
        string[] ss = p.Split(',');
        return new BoundingBox(double.Parse(ss[0]), double.Parse(ss[1]), double.Parse(ss[2]), double.Parse(ss[3]));
    }
}
